# -*- coding: utf-8 -*-
import logging

from pyramid.httpexceptions import HTTPFound

log = logging.getLogger(__name__)


SETTINGS_PREFIX = 'security.google_login.'


class AuthFailed(Exception):
    pass


def includeme(config):
    log.info("Add pyramid_google_login")

    config.include('pyramid_mako')
    config.include('.views')


def redirect_to_signin(request, message=None, url=None, headers=None):
    """ Redirect to the sign in page with message and next url """
    query = {}
    if message is not None:
        query['message'] = message
    if url is not None:
        query['url'] = url
    url = request.route_url('auth_signin', _query=query)
    return HTTPFound(location=url, headers=headers)


def find_landing_path(request):
    settings = request.registry.settings

    landing_url = settings.get(SETTINGS_PREFIX + 'landing_url')
    if landing_url is not None:
        return landing_url

    landing_route = settings.get(SETTINGS_PREFIX + 'landing_route')
    if landing_route is not None:
        try:
            return request.route_path(landing_route)
        except KeyError:
            pass
        try:
            return request.static_path(landing_route)
        except KeyError:
            pass

    return '/'



import unittest

import mock


class TestIncludeme(unittest.TestCase):

    def test_includeme(self):
        from pyramid_google_login import includeme

        config = mock.Mock()

        includeme(config)

        config.include.assert_has_calls([
            mock.call('pyramid_mako'),
            mock.call('.views'),
        ])


class TestHelpers(unittest.TestCase):

    def test_redirect_to_signin(self):
        from pyramid.httpexceptions import HTTPFound
        from pyramid_google_login import redirect_to_signin

        request = mock.Mock()

        httpfound = redirect_to_signin(request)
        self.assertIsInstance(httpfound, HTTPFound)

        request.route_url.assert_called_once_with('auth_signin', _query={})
        self.assertEqual(httpfound.location, request.route_url.return_value)

    def test_redirect_to_signin_url(self):
        from pyramid.httpexceptions import HTTPFound
        from pyramid_google_login import redirect_to_signin

        request = mock.Mock()

        httpfound = redirect_to_signin(request, url='/test')
        self.assertIsInstance(httpfound, HTTPFound)

        request.route_url.assert_called_once_with('auth_signin',
                                                  _query={'url': '/test'})
        self.assertEqual(httpfound.location, request.route_url.return_value)

    def test_redirect_to_signin_headers(self):
        from pyramid.httpexceptions import HTTPFound
        from pyramid_google_login import redirect_to_signin

        request = mock.Mock()

        test_header = ('X-Test', 'Yeap')

        httpfound = redirect_to_signin(request, headers=[test_header])
        self.assertIsInstance(httpfound, HTTPFound)

        self.assertEqual(httpfound.location, request.route_url.return_value)
        self.assertIn(test_header, httpfound.headerlist)

    def test_find_landing_path(self):
        from pyramid_google_login import find_landing_path

        request = mock.Mock()
        request.registry.settings = {

        }

        path = find_landing_path(request)

        self.assertEqual(path, '/')

    def test_find_landing_path_landing_url(self):
        from pyramid_google_login import find_landing_path

        request = mock.Mock()
        request.registry.settings = {
            'security.google_login.landing_url': '/foobar',
        }

        path = find_landing_path(request)

        self.assertEqual(path, '/foobar')

    def test_find_landing_path_landing_route(self):
        from pyramid_google_login import find_landing_path

        request = mock.Mock()
        request.registry.settings = {
            'security.google_login.landing_route': 'myroute',
        }

        path = find_landing_path(request)

        self.assertEqual(path, request.route_path.return_value)

    def test_find_landing_path_landing_route_static(self):
        from pyramid_google_login import find_landing_path

        request = mock.Mock()
        request.registry.settings = {
            'security.google_login.landing_route': 'myroute',
        }
        request.route_path.side_effect = KeyError()

        path = find_landing_path(request)

        self.assertEqual(path, request.static_path.return_value)

    def test_find_landing_path_fallback(self):
        from pyramid_google_login import find_landing_path

        request = mock.Mock()
        request.registry.settings = {
            'security.google_login.landing_route': 'myroute',
        }
        request.route_path.side_effect = KeyError()
        request.static_path.side_effect = KeyError()

        path = find_landing_path(request)

        self.assertEqual(path, '/')



import logging

from pyramid.view import view_config
# from pyramid.settings import asbool
from pyramid.security import (remember, forget, NO_PERMISSION_REQUIRED)
from pyramid.httpexceptions import HTTPFound

from pyramid_google_login import SETTINGS_PREFIX
from pyramid_google_login import redirect_to_signin, find_landing_path
from pyramid_google_login import AuthFailed
from pyramid_google_login.events import UserLoggedIn, UserLoggedOut
from pyramid_google_login.google_oauth2 import (build_authorize_url,
                                                exchange_token_from_code,
                                                get_userinfo_from_token,
                                                check_hosted_domain_user,
                                                get_user_id_from_userinfo,
                                                encode_state,
                                                decode_state,
                                                )

log = logging.getLogger(__name__)


def includeme(config):
    config.add_route('auth_signin', '/auth/signin')
    config.add_route('auth_signin_redirect', '/auth/signin_redirect')
    config.add_route('auth_callback', '/auth/oauth2callback')
    config.add_route('auth_logout', '/auth/logout')

    config.add_static_view('static/pyramid_google_login',
                           'pyramid_google_login:static',
                           cache_max_age=300)

    config.scan(__name__)


@view_config(route_name='auth_signin',
             permission=NO_PERMISSION_REQUIRED,
             renderer='pyramid_google_login:templates/signin.mako')
def signin(request):
    settings = request.registry.settings
    signin_banner = settings.get(SETTINGS_PREFIX + 'signin_banner')
    hosted_domain = settings.get(SETTINGS_PREFIX + 'hosted_domain')
    signin_advice = settings.get(SETTINGS_PREFIX + 'signin_advice')
    message = request.params.get('message')
    url = request.params.get('url')

    if request.authenticated_userid:
        if url:
            return HTTPFound(location=url)
        else:
            return HTTPFound(location=find_landing_path(request))

    if url:
        redirect_url = request.route_url('auth_signin_redirect',
                                         _query={'url': url})
    else:
        redirect_url = request.route_url('auth_signin_redirect')

    return {'signin_redirect_url': redirect_url,
            'message': message,
            'signin_banner': signin_banner,
            'signin_advice': signin_advice,
            'hosted_domain': hosted_domain,
            }


@view_config(route_name='auth_signin_redirect',
             permission=NO_PERMISSION_REQUIRED,)
def signin_redirect(request):
    state_params = {}
    if 'url' in request.params:
        state_params['url'] = request.params['url']

    state = encode_state(state_params)
    try:
        redirect_uri = build_authorize_url(request, state)
    except AuthFailed as err:
        log.warning("Google Login failed (%s)", err)
        return redirect_to_signin(request, "Google Login failed (%s)" % err)

    return HTTPFound(location=redirect_uri)


@view_config(route_name='auth_callback',
             permission=NO_PERMISSION_REQUIRED)
def callback(request):
    try:
        oauth2_token = exchange_token_from_code(request)
        userinfo = get_userinfo_from_token(oauth2_token)
        check_hosted_domain_user(request, userinfo)
        userid = get_user_id_from_userinfo(request, userinfo)

    except AuthFailed as err:
        log.warning("Google Login failed (%s)", err)
        return redirect_to_signin(request, "Google Login failed (%s)" % err)

    except Exception as err:
        log.warning("Google Login failed (%s)", err)
        # Protect against leaking critical information like client_secret
        return redirect_to_signin(request, "Google Login failed (unkown)")

    # Find the redirect url (fail-safe, the authentication is more important)
    try:
        state_params = decode_state(request.params['state'])
        url = state_params['url'][0]
    except:
        url = find_landing_path(request)

    user_logged_in = UserLoggedIn(request, userid, oauth2_token, userinfo)
    try:
        request.registry.notify(user_logged_in)
    except:
        log.exception("Application crashed processing UserLoggedIn event"
                      "\nuserinfo=%s oauth2_token=%s",
                      userinfo, oauth2_token)
        return redirect_to_signin(request,
                                  "Google Login failed (application error)")

    if user_logged_in.headers:
        headers = user_logged_in.headers
    else:
        headers = remember(request, principal=userid)
    return HTTPFound(location=url, headers=headers)


@view_config(route_name='auth_logout')
def logout(request):
    userid = request.unauthenticated_userid
    if userid is not None:
        event = UserLoggedOut(userid)
        request.registry.notify(event)

    headers = forget(request)
    return redirect_to_signin(request, "You are logged out!", headers=headers)



import unittest

from pyramid import testing
import mock


class TestIncludeme(unittest.TestCase):

    def test_includeme(self):
        from pyramid_google_login.views import includeme

        config = mock.Mock()

        includeme(config)

        expected_calls = [
            mock.call('auth_signin', '/auth/signin'),
            mock.call('auth_signin_redirect', '/auth/signin_redirect'),
            mock.call('auth_callback', '/auth/oauth2callback'),
            mock.call('auth_logout', '/auth/logout'),
        ]
        config.add_route.assert_has_calls(expected_calls, any_order=True)
        config.add_static_view.assert_called_once_with(
            'static/pyramid_google_login',
            'pyramid_google_login:static',
            cache_max_age=300)
        config.scan.assert_called_once_with('pyramid_google_login.views')


@mock.patch('pyramid_google_login.views.exchange_token_from_code')
@mock.patch('pyramid_google_login.views.get_userinfo_from_token')
@mock.patch('pyramid_google_login.views.get_user_id_from_userinfo')
@mock.patch('pyramid_google_login.views.decode_state')
@mock.patch('pyramid_google_login.views.remember')
class TestCallback(unittest.TestCase):

    def setUp(self):
        self.request = testing.DummyRequest()
        self.request.params = {
            'code': 'CODE1234',
            'state': 'STATE1234'
        }
        self.settings = {
            'security.google_login.client_id': 'CLIENTID',
            'security.google_login.client_secret': 'CLIENTSECRET',
        }

        testing.setUp(settings=self.settings)
        #self.request.route_url.return_value = '/test/url'

    def tearDown(self):
        testing.tearDown()

    def test_nominal(self, m_rem, m_dec_s, m_get_uid, m_get_u, m_exch_t):
        from pyramid_google_login.views import callback
        from pyramid.httpexceptions import HTTPFound

        m_rem.return_value = [('Set-Cookie', 'auth_tkt=plop')]
        m_dec_s.return_value = {'url': ['/next/url']}

        resp = callback(self.request)

        m_exch_t.assert_called_once_with(self.request)
        m_get_u.assert_called_once_with(m_exch_t.return_value)
        m_get_uid.assert_called_once_with(self.request, m_get_u.return_value)
        m_rem.assert_called_once_with(self.request,
                                      principal=m_get_uid.return_value)

        self.assertIsInstance(resp, HTTPFound)
        self.assertEqual(resp.location, '/next/url')

        expected_header = ('Set-Cookie', 'auth_tkt=plop')
        self.assertIn(expected_header, resp.headerlist)

    def test_event(self, m_rem, m_dec_s, m_get_uid, m_get_u, m_exch_t):
        from pyramid_google_login.views import callback
        from pyramid_google_login.events import UserLoggedIn

        subscriber = mock.Mock()

        with testing.testConfig() as config:
            config.add_subscriber(subscriber, UserLoggedIn)
            callback(self.request)

        subscriber.assert_called_once_with(mock.ANY)

        event = subscriber.call_args[0][0]

        self.assertEqual(event.userid, m_get_uid.return_value)
        self.assertEqual(event.oauth2_token, m_exch_t.return_value)
        self.assertEqual(event.userinfo, m_get_u.return_value)

    def test_event_error(self, m_rem, m_dec_s, m_get_uid, m_get_u, m_exch_t):
        from pyramid_google_login.views import callback
        from pyramid.httpexceptions import HTTPFound
        from pyramid_google_login.events import UserLoggedIn

        subscriber = mock.Mock()
        subscriber.side_effect = Exception()

        with testing.testConfig() as config:
            config.add_subscriber(subscriber, UserLoggedIn)
            config.add_route('auth_signin', '/test/url')
            resp = callback(self.request)

        self.assertIsInstance(resp, HTTPFound)

        self.assertEqual(
            resp.location,
            'http://example.com/test/url?message='
            'Google+Login+failed+%28application+error%29',
            )

    def test_no_url(self, m_rem, m_dec_s, m_get_uid, m_get_u, m_exch_t):
        from pyramid_google_login.views import callback
        from pyramid.httpexceptions import HTTPFound

        m_rem.return_value = [('Set-Cookie', 'auth_tkt=plop')]
        m_dec_s.return_value = {}

        resp = callback(self.request)

        self.assertIsInstance(resp, HTTPFound)
        self.assertEqual(resp.location, '/')

    def test_oauth2_flow_error(self, m_rem, m_dec_s, m_get_uid, m_get_u,
                               m_exch_t):
        from pyramid_google_login.views import callback
        from pyramid_google_login import AuthFailed

        from pyramid.httpexceptions import HTTPFound

        m_exch_t.side_effect = AuthFailed('TESTEXC')

        with testing.testConfig() as config:
            config.add_route('auth_signin', '/test/url')
            test_exc = callback(self.request)

        self.assertIsInstance(test_exc, HTTPFound)

        self.assertEqual(
            test_exc.location,
            'http://example.com/test/url?message='
            'Google+Login+failed+%28TESTEXC%29',
            )

    def test_unkown_error(self, m_rem, m_dec_s, m_get_uid, m_get_u,
                          m_exch_t):
        from pyramid.httpexceptions import HTTPFound
        from pyramid_google_login.views import callback

        m_exch_t.side_effect = Exception('TESTEXC')

        with testing.testConfig() as config:
            config.add_route('auth_signin', '/test/url')
            test_exc = callback(self.request)

        self.assertIsInstance(test_exc, HTTPFound)

        self.assertEqual(
            test_exc.exception.location,
            'http://example.com/test/url?message='
            'Google+Login+failed+%28unkown%29',
            )

    def test_custom_cookie(self, m_rem, m_dec_s, m_get_uid, m_get_u,
                           m_exch_t):
        from pyramid_google_login.views import callback
        from pyramid_google_login.events import UserLoggedIn

        m_dec_s.return_value = {'url': ['/next/url']}

        def subscriber(event):
            event.headers = [('Set-Cookie', 'user=bob')]

        with testing.testConfig() as config:
            config.add_subscriber(subscriber, UserLoggedIn)
            response = callback(self.request)

        self.assertIn('Set-Cookie', response.headers)
        self.assertEqual('user=bob', response.headers.get('Set-Cookie'))


@mock.patch('pyramid_google_login.views.find_landing_path')
class TestSignin(unittest.TestCase):

    def setUp(self):
        self.request = mock.Mock()
        self.request.params = {
        }
        self.settings = {
            'security.google_login.signin_banner': 'CLIENTID',
            'security.google_login.client_secret': 'CLIENTSECRET',
        }
        self.request.registry.settings = self.settings

        self.request.authenticated_userid = None
        self.request.route_url.return_value = '/test/url'

    def test_nominal(self, m_find_landing_path):
        from pyramid_google_login.views import signin
        # from pyramid.httpexceptions import HTTPFound

        resp = signin(self.request)

        expected = {'hosted_domain': None,
                    'message': None,
                    'signin_advice': None,
                    'signin_banner': 'CLIENTID',
                    'signin_redirect_url': '/test/url'}

        self.assertEqual(resp, expected)

    def test_authenticated(self, m_find_landing_path):
        from pyramid_google_login.views import signin
        from pyramid.httpexceptions import HTTPFound

        self.request.authenticated_userid = '123'

        resp = signin(self.request)

        self.assertIsInstance(resp, HTTPFound)
        self.assertEqual(resp.location, m_find_landing_path.return_value)

    def test_authenticated_with_url(self, m_find_landing_path):
        from pyramid_google_login.views import signin
        from pyramid.httpexceptions import HTTPFound

        self.request.params['url'] = '/go/there'
        self.request.authenticated_userid = '123'

        resp = signin(self.request)

        self.assertIsInstance(resp, HTTPFound)
        self.assertEqual(resp.location, '/go/there')


@mock.patch('pyramid_google_login.views.redirect_to_signin')
class TestLogout(unittest.TestCase):

    def setUp(self):
        self.request = mock.Mock()
        self.settings = {}
        self.request.registry.settings = self.settings

    def test_nominal(self, m_redirect):
        from pyramid_google_login.views import logout
        from pyramid_google_login.events import UserLoggedOut, Event

        resp = logout(self.request)

        self.assertEqual(resp, m_redirect.return_value)

        self.assertEqual(self.request.registry.notify.called, True)
        event = self.request.registry.notify.call_args[0][0]

        self.assertIsInstance(event, UserLoggedOut)
        self.assertIsInstance(event, Event)
        self.assertEqual(event.userid, self.request.unauthenticated_userid)
